import numpy as np
import pickle
from sklearn.linear_model import LinearRegression as LR
from sklearn.linear_model import LassoCV
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestRegressor as RF
from sklearn.ensemble import GradientBoostingRegressor as GB
from sklearn import metrics
from sklearn.cluster import SpectralClustering as SC
from sklearn import preprocessing
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from typing import List, Dict, Tuple
from method import *
import time
import sys

'''
    plotted:
        clustering accuracy versus SNR (in terms of coeff)
        clustering accuracy versus p (in terms of coeff) 

'''

# global param
# snr=p/s^2 assuming p standard Gaussian and sd=s noise
# snr = np.power(10, np.arange(0,3,step=0.5))
snr = np.power(2, np.arange(0,8,step=1, dtype=float))
print('snr: ', snr)
n = 50
p = 5 #5, 10, 15
print('p: ', p)
numClut = 2 # num of clusters
numModu = 10 # num of modules per cluster
nrep = 100 # num of rep
sds = np.sqrt(1.0 * p / snr) # sd of noise
ns = len(sds)
m = numModu * numClut # num of modules


def genDat(b, s):
    v = np.random.normal(size=n) * s
    X = np.random.normal(size=(n, p))
    y = X.dot(b) + v

    return X, y


def check(ks, labs):
    '''
        AD HOC function, only for 
        numClut = 2 case

        check whether the clustering is accurate

    '''
    nrep = len(ks) #num of rep
    acc = 0
    for r in range(nrep):
        if ks[r] == 2 and np.abs(labs[r][:numModu].sum() - labs[r][numModu:].sum())==numModu:
            # the second condition holds only if the first c labels are 0 and last c are 1
            # or vice versu
            acc += 100.0 / nrep
    # print('acc: ', acc)
    return acc


def run(s):

    # true label of cluster
    lab0 = []

    # gen data for each module
    for i in range(numClut):
        b = np.random.normal(size=p)
        for j in range(numModu):
            lab0.append(i)
            X, y = genDat(b, s)
            # X = preprocessing.scale(X)
            # y = preprocessing.scale(y)
            Xs.append(X)
            ys.append(y)
            reg0 = RF(n_estimators=50, max_depth=3)
            #reg = GB(n_estimators=100, max_depth=5)
            reg1 = LassoCV(cv=2, random_state=0)
            scores = cross_val_score(reg0, X, y, cv=5)
            scores1 = cross_val_score(reg1, X, y, cv=5)
            # print('np.mean(scores): ', np.mean(scores))
            if np.mean(scores) > np.mean(scores1):
                regs.append(reg0)
            else:
                regs.append(reg1)
            regs[-1].fit(X, y)
            # print('len(regs) ', len(regs))

    # validation error matrix
    V = np.zeros((m, m))
    for i in range(m):
        for j in range(m):
            # print('i, j', i, j)
            y_pred = regs[i].predict(Xs[j])
            # Mean absolute error regression loss
            # validation of i on data j
            # V[i, j] = metrics.mean_absolute_error(ys[j], y_pred)
            # Mean squared error regression loss
            V[i, j] = metrics.mean_squared_error(ys[j], y_pred) 

    # weight matrix
    # not yet normalize the weight to sum to one
    W = np.zeros((m, m))
    for i in range(m):
        for j in range(m):
            #aij= ns[j]*(V[j, i] + V[j, j])
            #bij= ns[i]*(V[j, i] + V[j, j])
            #aaij=aij/(aij+bij)
            #W[i, j] = aaij*np.abs(V[i, j] - V[j, j]) + (1-aaij)*np.abs(V[j, i] - V[i, i])
            W[i, j] = np.abs(V[i, j] - V[j, j]) + np.abs(V[j, i] - V[i, i])
    W = np.exp(-W)  # transform to 'SIMILARITY'

    # print('\nV = \n', np.round(V, 2))
    # print('\nW = \n', np.round(W, 2))

    # getClust(W, 3)  # this function is not in use
    lab, k = getClust_custom(W, k=None, PRINT=True)

    return lab, k, lab0


if __name__ == '__main__':

    start = time.time()
    print("start running")
    count = 0

    accs = [] #acc for each s
    for i in range(ns):
        s = sds[3]

        labs=[]
        ks=[]

        for r in range(nrep):

            Xs = []
            ys = []
            regs = []
            
            # print('replication ', r)
            lab, k, lab0 = run(s)
            labs.append(lab)
            ks.append(k)

            count += 1
            t = np.round(time.time() - start, 2)
            print('elapsed time (s): ', t)
            print('estimated overall time (s): ', int(t * (ns * nrep / count)))

        print('ks', ks)
        # print('labs', labs)
        acc = check(ks, labs)
        accs.append(acc)
        print('\naccs: ', accs)

    print('\naccs: ', accs)
            
    print('\ntrue labels are: \n', lab0)

